#define std_io_OutputTube_IMPORT
#define std_io_BufferedOutputTube_IMPORT
#define std_lang_Object_IMPORT

#include <stdlib.h>
#include <string.h>
#include <std/io/impl/BufferedOutputTube.h>

union BOT {
    struct impl_OutputTube *outputTube;
    struct impl_BufferedOutputTube *impl;
} __attribute__((__transparent_union__));

static int
flush(struct impl_BufferedOutputTube *impl)
{
    int s;

    s = OutputTube_writeFully(impl->dest, impl->bufferData, impl->size);
    impl->size = 0;
    return s;
}

static ssize_t
BOT_write(union BOT this, const void *data, size_t size)
{
    struct impl_BufferedOutputTube *impl = this.impl;
    union OutputTube dest = impl->dest;
    size_t delta;

    if (impl->size == 0 && size > impl->bufferSize) {
	return ((OutputTube_writeFully(dest, data, size) < 0)
		? -1 : (ssize_t)size);
    }
    delta = impl->bufferSize - impl->size;
    if (size > delta)
	size = delta;
    memcpy((char *)impl->bufferData + impl->size, data, size);
    impl->size += size;
    return ((impl->size == impl->bufferSize && flush(impl) < 0)
	    ? -1 : (ssize_t)size);
}

static int
BOT_close(union BOT this)
{
    struct impl_BufferedOutputTube *impl = this.impl;
    union OutputTube dest = impl->dest;
    int s;

    s = (impl->size == 0) ? 0 : flush(impl);
    s = (OutputTube_close(dest) < 0) ? -1 : s;
    free(impl->bufferData);
    free(impl);
    return s;
}

static const struct OutputTube_Methods BOT_METHODS = {
    .write = BOT_write,
    .close = BOT_close};

union BufferedOutputTube
BufferedOutputTube_create(union OutputTube dest)
{
    struct impl_BufferedOutputTube *impl;
    static const size_t SIZE = 4096;

    if (dest.impl == NULL) {
	return (union BufferedOutputTube)Object_NULL;
    }
    if ((impl = malloc(sizeof(*impl))) == NULL) {
	OutputTube_close(dest);
        return (union BufferedOutputTube)Object_NULL;
    }
    OutputTube_initialize(&impl->super.outputTube, &BOT_METHODS);
    impl->dest = dest;
    impl->size = 0;
    impl->bufferSize = SIZE;
    impl->bufferData = NULL;

    if ((impl->bufferData = malloc(SIZE)) == NULL) {
	BOT_close(impl);
	return (union BufferedOutputTube)Object_NULL;
    }
    return (union BufferedOutputTube)impl;
}
