#define std_lang_Object_IMPORT
#define std_util_LinkedHeap_IMPORT

#include <stdlib.h>
#include <std/util/impl/LinkedHeap.h>

static struct LinkedHeap_Node *
LinkedHeap_Node_create(void *c)
{
    struct LinkedHeap_Node *n;
    
    if ((n = (struct LinkedHeap_Node *)malloc(sizeof(*n))) == NULL)
        return NULL;
    n->next = NULL;
    n->prev = NULL;
    n->parent = NULL;
    n->left = NULL;
    n->right = NULL;
    n->value = c;
    return n;
}

static void
LinkedHeap_Node_free(struct LinkedHeap_Node *n)
{
    free(n);
}

union LinkedHeap
LinkedHeap_create(int (*compare)(void *, void *))
{
    struct impl_LinkedHeap *impl;

    if ((impl = (struct impl_LinkedHeap *)malloc(sizeof(*impl))) == NULL) {
	return (union LinkedHeap)Object_NULL;
    }
    impl->root = NULL;
    impl->last = NULL;
    impl->nodes = 0;
    impl->compare = compare;
    return (union LinkedHeap)impl;
}

void
LinkedHeap_free(union LinkedHeap heap)
{
    struct impl_LinkedHeap *impl = heap.impl;
    struct LinkedHeap_Node *node = impl->root, *next;

    while (node != NULL) {
	next = node->next;
	LinkedHeap_Node_free(node);
	node = next;
    }
    free(impl);
}

#define COMPARE(h, p, q) ((h)->compare((p)->value, (q)->value))

static void
swap(struct LinkedHeap_Node *p, struct LinkedHeap_Node *q)
{
    void *c;

    c = p->value;
    p->value = q->value;
    q->value = c;
}

static void
reconfigureAtTail(struct impl_LinkedHeap *h)
{
    struct LinkedHeap_Node *n, *p;

    for (n = h->last; (p = n->parent) != NULL && COMPARE(h, n, p) < 0; n = p)
	swap(n, p);
}

static void
add(struct impl_LinkedHeap *h, struct LinkedHeap_Node *n)
{
    struct LinkedHeap_Node *parent = h->last->parent;

    if (parent == NULL) {
	/* LNοƤNULL => LNϥ롼,  IPϥ롼Ȥκλ */
	parent = h->root;
	parent->left = n;
    }
    else if (parent->right == NULL) {
	/* LNοƤαλҤNULL => IPοƤLNοơIPLNȷ */
	parent->right = n;
    }
    else {
        /* LNοƤϻҤ2Ĥ => IPοƤLNοƤΡּפΥΡ */
	parent = parent->next;
	parent->left = n;
    }
    n->prev = h->last;
    n->parent = parent;
    h->last->next = n;
    h->last = n;
}

int
LinkedHeap_put(union LinkedHeap heap, void *c)
{
    struct impl_LinkedHeap *h = heap.impl;
    struct LinkedHeap_Node *n;

    if ((n = LinkedHeap_Node_create(c)) == NULL) {
	return -1;
    }
    if (h->root == NULL) {
	h->root = n;
	h->last = n;
	h->nodes = 1;
	return 0;
    }
    add(h, n);
    reconfigureAtTail(h);
    ++(h->nodes);
    return 0;
}

static struct LinkedHeap_Node *
minChild(struct impl_LinkedHeap *h, struct LinkedHeap_Node *n)
{
    if (n->left == NULL)
	return NULL;
    if (n->right == NULL || COMPARE(h, n->left, n->right) < 0)
	return n->left;
    return n->right;
}

static void
reconfigureAtRoot(struct impl_LinkedHeap *h)
{
    struct LinkedHeap_Node *n, *c;

    for (n = h->root; (c = minChild(h, n)) != NULL
	     && COMPARE(h, n, c) > 0; n = c) {
	swap(n, c);
    }
}

static void
deleteLast(struct impl_LinkedHeap *h)
{
    struct LinkedHeap_Node *parent, *last;
    
    parent = h->last->parent;
    last = h->last->prev;
    if (parent == NULL) {
	/* LNοƤNULL => LNϥ롼 */
	h->root = NULL;
	/* ΤȤ last == NULL */
    }
    else if (parent->right == h->last) {
	/* LNοƤαλҤLN */
	parent->right = NULL;
	last->next = NULL;
    }
    else {
	/* LNοƤκλҤLN */
	parent->left = NULL;
	last->next = NULL;
    }
    LinkedHeap_Node_free(h->last);
    h->last = last;
    --(h->nodes);
}

void *
LinkedHeap_getFirst(union LinkedHeap heap)
{
    struct impl_LinkedHeap *h = heap.impl;
    void *value;

    if (h->root == NULL)
	return NULL;
    value = h->root->value;
    h->root->value = h->last->value;
    deleteLast(h);
    if (h->nodes > 0) {
	reconfigureAtRoot(h);
    }
    return value;
}

size_t
LinkedHeap_getSize(union LinkedHeap heap)
{
    struct impl_LinkedHeap *h = heap.impl;

    return h->nodes;
}


void
LinkedHeap_forEach(union LinkedHeap heap, void (*callback)(void *))
{
    struct impl_LinkedHeap *h = heap.impl;
    struct LinkedHeap_Node *n;

    for (n = h->root; n != NULL; n = n->next) {
	callback(n->value);
    }
}
